"""

USAGE:
    Format:
        format1
            clpaws vipctl --replace-route -v VPC_ID
                                          -i IP -n EC2_NETWORK_IF_ID
        format2
            clpaws vipctl --check-subnet -v VPC_ID -i IP
        format3
            clpaws vipctl --check-route -v VPC_ID
                                        -i IP -n EC2_NETWORK_IF_ID

    Argument:
        --replace-route      : replace to Route Table
        --check-subnet       : check VPC subnet
        --check-route        : check Route Table of Virtual IP Address
        --debug              : output debug log (optional)
        -i IP                : Virtual IP Address
        -n EC2_NETWORK_IF_ID : EC2 Network Interface ID
        -v VPC_ID            : VPC ID
        -t TIMEOUT           : timeout (optional)

    ReturnCode:
        value of clpaws.errors.ERRORNO

    Example:
        example1
            clpaws vipctl --replace-route -v vpc-00000000
                          -i 10.1.0.20 -n eni-00000000
        example2
            clpaws vipctl --check-subnet -v vpc-00000000
                          -i 10.1.0.20
        example3
            clpaws vipctl --check-route -v vpc-00000000
                          -i 10.1.0.20 -n eni-00000000

"""


import sys
import optparse
import clpaws.utils as util
import clpaws.errors as error
import logging
import os
import socket
import struct

# global varivable
loglevel = logging.WARNING

def _check_option():
    exclusive_options = ('replace_route','check_subnet', 'check_route')

    parser = optparse.OptionParser()

    # exclusive long options
    parser.add_option('--replace-route',
                       action='store_true',
                       dest='replace_route',
                       help='replace Route Table')
    parser.add_option('--check-subnet',
                       action='store_true',
                       dest='check_subnet',
                       help='check VPC subnet')
    parser.add_option('--check-route',
                        dest='check_route',
                        action='store_true',
                        help='check Route Table of Virtual IP Address')

    # optional long option
    parser.add_option('--debug',
                        dest='debug',
                        action='store_true',
                        help='output debug log')

    # required short options
    parser.add_option('-i',
                        dest='ip',
                        help='Virtual IP Address',
                        metavar='IP')
    parser.add_option('-n',
                        dest='ec2_network_if_id',
                        help='EC2 Network Interface ID',
                        metavar='EC2_NETWORK_IF_ID')
    parser.add_option('-v',
                        dest='vpc_id',
                        help='VPC ID',
                        metavar='VPC_ID')

    # optional short option
    parser.add_option('-t',
                        dest='timeout',
                        help='timeout',
                        metavar='TIMEOUT')

    (options, args) = parser.parse_args()

    toggled_options = [name for name in exclusive_options
                        if getattr(options, name)]

    if not len(toggled_options) == 1:
        parser.error('only one argument is required '
                     'among arguments "--check-subnet", "--replace-route" and '
                     ' "--check-route".')

    if options.replace_route and not (options.ip and options.ec2_network_if_id
                                        and options.vpc_id):
        parser.error('3 arguments "-v", "-i" and "-n" are required '
                     'when argument "--replace-route" is specified.')

    if options.check_subnet and not (options.ip and options.vpc_id):
        parser.error('2 arguments "-v" and "-i" are required '
                     'when argument "--check-subnet" is specified.')

    if  options.check_route and not (options.ip and options.ec2_network_if_id
                                        and options.vpc_id):
        parser.error('3 arguments "-v", "-i" and "-n" are required '
                     'when argument "--check-route" is specified.')

    return options


def check_subnet(ip_address, vpc_id, timeout):
    logging.debug('Start check_subnet.')

    vpc_cidrblock = _get_vpc_cidrblock(vpc_id, timeout)
    vpc_network_address, vpc_netmask_length = vpc_cidrblock.split('/')

    vpc_network_address_bit  = \
        _get_network_addrss_bits(vpc_network_address, vpc_netmask_length)
    network_address_bit = \
        _get_network_addrss_bits(ip_address, vpc_netmask_length)

    if vpc_network_address_bit == network_address_bit:
        err_msg = 'VIP {0} is belong to subnet of VPC {1}.'.format(ip_address,
                                                                        vpc_id)
        raise error.AWSCLIError(error.ERRORNO['check_subnet_error'],
                                err_msg)

    logging.debug('End check_subnet.')


def replace_route_table(route_table_id, ip_address, ec2_network_if_id, timeout):
    logging.debug('Start replace_route_table.')

    cidr_block = ip_address + '/32'

    cmdline = ('aws ec2 replace-route --route-table-id "{0}" '
               '--destination-cidr-block "{1}" --network-interface-id "{2}"'
               .format(route_table_id, cidr_block, ec2_network_if_id))

    util.execute_awscli(cmdline, timeout,
                        err_msg='Route table ({0}) failed to replace.'
                        .format(route_table_id))

    logging.debug('End replace_route_table.')


def check_route_table(vpc_id, ip_address, ec2_interface_id, timeout):
    logging.debug('Start check_route_table.')

    changed_route_tabel_id_list = \
        _get_changed_route_table_id_list(vpc_id, ip_address,
                                            ec2_interface_id, timeout)
    if not (len(changed_route_tabel_id_list) == 0):
        err_msg = 'VIP {0} route of {1} is changed.'.format(ip_address,
                                                    changed_route_tabel_id_list)
        raise error.AWSCLIError(error.ERRORNO['check_route_table_error'],
                                err_msg)

    logging.debug('End check_route_table.')


# Get IDs of route tables which have VIP's routing
def _get_route_table_id_list(vpc_id, ip_address, timeout):
    cidr_block = ip_address + '/32'

    cmdline = ('aws ec2 describe-route-tables '
                '--filters Name=vpc-id,Values="{0}"'.format(vpc_id))
    result = util.execute_awscli(cmdline, timeout)
    route_tables = result.get('RouteTables')

    if len(route_tables) == 0:
        logging.warning('Route table of VPC {0} does not exist.'.format(vpc_id))

    route_table_id_list = []
    for i in route_tables:
        routes = i.get('Routes')
        exist_vip_routes = False

        for j in routes:
            destination_cidr_block = j.get('DestinationCidrBlock')

            if destination_cidr_block == cidr_block:
                exist_vip_routes = True
                break

        if exist_vip_routes == True:
            route_table_id_list.append(i.get('RouteTableId'))

    return route_table_id_list


def _get_changed_route_table_id_list(vpc_id, ip_address,
                                        ec2_interface_id, timeout):
    cidr_block = ip_address + '/32'

    cmdline = ('aws ec2 describe-route-tables '
                '--filters Name=vpc-id,Values={0}'.format(vpc_id))

    result = util.execute_awscli(cmdline, timeout)
    route_tables = result.get('RouteTables')

    if len(route_tables) == 0:
        logging.warning('Route table of VPC {0} does not exist.'.format(vpc_id))

    changed_route_tabel_id_list = []

    for i in route_tables:
        routes = i.get('Routes')
        change_ec2_id = False

        for j in routes:
            destination_cidr_block = j.get('DestinationCidrBlock')
            network_interface_id = j.get('NetworkInterfaceId')

            if destination_cidr_block == cidr_block:
                if network_interface_id != ec2_interface_id:
                     change_ec2_id = True
                     break

        if change_ec2_id == True:
            changed_route_tabel_id_list.append(i.get('RouteTableId'))

    return changed_route_tabel_id_list


def _get_vpc_cidrblock(vpc_id, timeout):
    cmdline = ('aws ec2 describe-vpcs --vpc-ids {0}'
                .format(vpc_id))

    result = util.execute_awscli(cmdline, timeout)
    vpc = result.get('Vpcs')

    if (vpc and len(vpc) == 1):
        vpc_cidrblock = vpc[0].get('CidrBlock')
        if vpc_cidrblock:
            return vpc_cidrblock
        else:
            err_msg = 'CidrBlock is invalid.'
            raise error.AWSCLIError(error.ERRORNO['internal_error'], err_msg)
    else:
        err_msg = 'Vpcs is invalid.'
        raise error.AWSCLIError(error.ERRORNO['internal_error'], err_msg)


def _get_network_addrss_bits(ip_address, netmask_length):
    ip_address_bit = \
        struct.unpack('>L', socket.inet_aton(ip_address))[0]
    netmask_bit = (0xffffffff << (32 - int(netmask_length)) & 0xffffffff)
    network_addres_bit = ip_address_bit & netmask_bit
    return network_addres_bit


def main():
    global loglevel
    try:
        options = _check_option()

        if options.debug:
            loglevel = logging.DEBUG

        logging.basicConfig(level=loglevel,
                            format='%(levelname)s - %(message)s',
                            stream=sys.stdout)

        logging.debug('Start clpaws vipctl.')

        util.set_environ()

        if options.replace_route:
            route_table_id_list = _get_route_table_id_list(options.vpc_id,
                                                            options.ip, options.timeout)
            logging.debug('Route tables which have VIP({0})\'s routing are {1}'
                            .format(options.ip, route_table_id_list))

            for route_table_id in route_table_id_list:
                replace_route_table(route_table_id, options.ip,
                                    options.ec2_network_if_id, options.timeout)
        elif options.check_subnet:
            check_subnet(options.ip, options.vpc_id, options.timeout)
        elif options.check_route:
            check_route_table(options.vpc_id, options.ip,
                                options.ec2_network_if_id, options.timeout)
        else:
            err_msg = 'Invalid option.'
            raise error.AWSCLIError(error.ERRORNO['internal_error'], err_msg)

        logging.debug('End clpaws vipctl.')
        return error.ERRORNO['success']

    except error.AWSCLIError as ex:
        util.handle_exception(ex)
        return ex.value

    except Exception as ex:
        logging.error(ex)
        return error.ERRORNO['internal_error']


if __name__ == '__main__':
    import sys
    sys.exit(main())
